@@ -5,12 +5,15 @@ import json
 import queue
 import secrets
 import threading
+import time
 import typing
+from types import TracebackType
 
 import anyio
 import httpcore
 import httpx
 import wsproto
+from anyio.streams.memory import MemoryObjectReceiveStream, MemoryObjectSendStream
 from httpcore import AsyncNetworkStream, NetworkStream
 from wsproto.frame_protocol import CloseReason
 
@@ -515,21 +518,32 @@ class WebSocketSession:
     def _wait_until_closed(
         self, callable: typing.Callable[..., TaskResult], *args, **kwargs
     ) -> TaskResult:
+        exit_await = threading.Event()
+
+        def wait_close() -> None:
+            while not exit_await.is_set() and not self._should_close.is_set():
+                time.sleep(0.05)
+
+        executor = concurrent.futures.ThreadPoolExecutor(max_workers=2)
         try:
-            executor = concurrent.futures.ThreadPoolExecutor(max_workers=2)
-            wait_close_task = executor.submit(self._should_close.wait)
+            wait_close_task = executor.submit(wait_close)
             todo_task = executor.submit(callable, *args, **kwargs)
         except RuntimeError as e:
             raise ShouldClose() from e
-        done, pending = concurrent.futures.wait(  # type: ignore
-            (todo_task, wait_close_task), return_when=concurrent.futures.FIRST_COMPLETED
-        )
-        for task in pending:
-            task.cancel()
-        if wait_close_task in done and todo_task not in done:
-            raise ShouldClose()
-        result = todo_task.result()
-        executor.shutdown(False)
+        else:
+            done, _ = concurrent.futures.wait(
+                (todo_task, wait_close_task),  # type: ignore[misc]
+                return_when=concurrent.futures.FIRST_COMPLETED,
+            )
+            if wait_close_task in done:
+                raise ShouldClose()
+            assert todo_task in done
+            if not wait_close_task.cancel():
+                exit_await.set()
+                wait_close_task.result()
+            result = todo_task.result()
+        finally:
+            executor.shutdown(False)
         return result
 
 
@@ -546,6 +560,12 @@ class AsyncWebSocketSession:
 
     subprotocol: typing.Optional[str]
     response: typing.Optional[httpx.Response]
+    _send_event: MemoryObjectSendStream[
+        typing.Union[wsproto.events.Event, HTTPXWSException]
+    ]
+    _receive_event: MemoryObjectReceiveStream[
+        typing.Union[wsproto.events.Event, HTTPXWSException]
+    ]
 
     def __init__(
         self,
@@ -569,10 +589,6 @@ class AsyncWebSocketSession:
         else:
             self.subprotocol = None
 
-        self._send_event, self._receive_event = anyio.create_memory_object_stream[
-            typing.Union[wsproto.events.Event, HTTPXWSException]
-        ]()
-
         self._ping_manager = AsyncPingManager()
         self._should_close = anyio.Event()
 
@@ -587,26 +603,39 @@ class AsyncWebSocketSession:
             self._keepalive_ping_interval_seconds = keepalive_ping_interval_seconds
             self._keepalive_ping_timeout_seconds = keepalive_ping_timeout_seconds
 
-    async def __aenter__(self):
-        self._exit_stack = contextlib.AsyncExitStack()
-        self._background_task_group = anyio.create_task_group()
-        await self._exit_stack.enter_async_context(self._background_task_group)
+    async def __aenter__(self) -> "AsyncWebSocketSession":
+        async with contextlib.AsyncExitStack() as exit_stack:
+            self._send_event, self._receive_event = anyio.create_memory_object_stream[
+                typing.Union[wsproto.events.Event, HTTPXWSException]
+            ]()
+            exit_stack.enter_context(self._send_event)
+            exit_stack.enter_context(self._receive_event)
+
+            self._background_task_group = anyio.create_task_group()
+            await exit_stack.enter_async_context(self._background_task_group)
 
-        self._background_task_group.start_soon(
-            self._background_receive, self._max_message_size_bytes
-        )
-        if self._keepalive_ping_interval_seconds is not None:
             self._background_task_group.start_soon(
-                self._background_keepalive_ping,
-                self._keepalive_ping_interval_seconds,
-                self._keepalive_ping_timeout_seconds,
+                self._background_receive, self._max_message_size_bytes
             )
+            if self._keepalive_ping_interval_seconds is not None:
+                self._background_task_group.start_soon(
+                    self._background_keepalive_ping,
+                    self._keepalive_ping_interval_seconds,
+                    self._keepalive_ping_timeout_seconds,
+                )
+
+            exit_stack.callback(self._background_task_group.cancel_scope.cancel)
+            exit_stack.push_async_callback(self.close)
+            self._exit_stack = exit_stack.pop_all()
 
         return self
 
-    async def __aexit__(self, exc_type, exc, tb):
-        await self.close()
-        self._background_task_group.cancel_scope.cancel()
+    async def __aexit__(
+        self,
+        exc_type: typing.Optional[typing.Type[BaseException]],
+        exc: typing.Optional[BaseException],
+        tb: typing.Optional[TracebackType],
+    ) -> None:
         await self._exit_stack.aclose()
 
     async def ping(self, payload: bytes = b"") -> anyio.Event:
@@ -989,27 +1018,22 @@ class AsyncWebSocketSession:
         except (httpcore.ReadError, httpcore.WriteError):
             await self.close(CloseReason.INTERNAL_ERROR, "Stream error")
             await self._send_event.send(WebSocketNetworkError())
-        except anyio.get_cancelled_exc_class():
-            pass
 
     async def _background_keepalive_ping(
         self, interval_seconds: float, timeout_seconds: typing.Optional[float] = None
     ) -> None:
-        try:
-            while not self._should_close.is_set():
-                await anyio.sleep(interval_seconds)
-                pong_callback = await self.ping()
-                if timeout_seconds is not None:
-                    try:
-                        with anyio.fail_after(timeout_seconds):
-                            await pong_callback.wait()
-                    except TimeoutError:
-                        await self.close(
-                            CloseReason.INTERNAL_ERROR, "Keepalive ping timeout"
-                        )
-                        await self._send_event.send(WebSocketNetworkError())
-        except anyio.get_cancelled_exc_class():
-            pass
+        while not self._should_close.is_set():
+            await anyio.sleep(interval_seconds)
+            pong_callback = await self.ping()
+            if timeout_seconds is not None:
+                try:
+                    with anyio.fail_after(timeout_seconds):
+                        await pong_callback.wait()
+                except TimeoutError:
+                    await self.close(
+                        CloseReason.INTERNAL_ERROR, "Keepalive ping timeout"
+                    )
+                    await self._send_event.send(WebSocketNetworkError())
 
 
 def _get_headers(
